# -*- coding: utf-8 -*-
#
# modulemd_summarizer - Prints a summary of ModuleMD files
#
# Copyright © 2018 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author:
# Rafael dos Santos <rdossant@redhat.com>

from collections import defaultdict
from fnmatch import fnmatch
import tempfile

from click import ClickException
import gi
gi.require_version('Modulemd', '1.0')  # noqa: E402
from gi.repository import Modulemd
import smartcols

from . import _fetchrepodata, _repodata


def _print_summary(profiles, sdefaults, pdefaults, deps, mfilter, as_tree):
    tb = smartcols.Table()
    cl = tb.new_column('Name')
    cl.tree = as_tree
    cl_strm = tb.new_column('Stream')
    cl_ver = tb.new_column('Version')
    cl_ctxt = tb.new_column('Context')
    cl_prof = tb.new_column('Profiles')
    cl_deps = tb.new_column('Dependencies')
    parent_ln = {}
    for nsvc, plist in sorted(profiles.items()):
        if mfilter and not any((fnmatch(nsvc, p) for p in mfilter)):
            continue

        modname, sname, version, context = nsvc.split(':')

        def is_def_strm(s):
            return s == sdefaults.get(modname, '')

        if as_tree:
            pl = parent_ln.get(modname, None)
            if pl is None:
                pl = parent_ln.setdefault(modname, tb.new_line())
            pl[cl] = modname
        else:
            pl = None

        ln = tb.new_line(pl)
        ln[cl] = modname
        ln[cl_strm] = sname + ' [d]' * is_def_strm(sname)
        ln[cl_ver] = version
        ln[cl_ctxt] = context

        def is_def_prof(p):
            return p in pdefaults.get(modname, {}).get(sname, [])

        ln[cl_prof] = ', '.join(p + ' [d]' if is_def_prof(p) else p
                                for p in plist)

        dlist = sorted(deps.get(nsvc, []))
        if len(dlist) == 0:
            continue
        if len(dlist) == 1:
            dlist = str(dlist[0])
        else:
            dlist = ','.join(dlist)
        ln[cl_deps] = dlist

    print(tb)
    print('\nHint: [d]efault')


def _parse_mmd(mmd_index, profiles, dstreams, dprofiles, deps):
    """
    Parse module repodata information. This repodata is provided by a tool
    (e.g MBS), so we can assume the information is well-formed
    """

    for index in mmd_index:
        module_name = index.get_name()
        for module in index.get_streams().values():
            nsvc = module.get_nsvc()
            plist = list(module.get_profiles().keys())
            profiles[nsvc] = sorted(set(profiles.get(nsvc, []) + plist))

            for dep in module.get_dependencies():
                deplist = set(deps.get(nsvc, []))
                for m, s in dep.peek_requires().items():
                    deplist.add(f"{m}:{','.join(s.get())}"
                                if len(s.get()) else m)
                deps[nsvc] = sorted(deplist)

        defaults = index.get_defaults()
        if not defaults:
            continue

        # Local module metadata can overwrite metadata from repo
        dstreams[module_name] = defaults.peek_default_stream()
        for s, pset in defaults.peek_profile_defaults().items():
            dprofiles[module_name][s] = pset.get()


def _parse_objs(objects, profiles, dstreams, dprofiles, deps):
    """
    Parse module metadata information. This metadata is probably provided by
    packager-written yamls, so it might be missing information (like module
    name or module defaults).
    """

    for obj in objects:
        if isinstance(obj, Modulemd.Module):
            # local modulemd files might miss some info like context or
            # version. So let's make sure nsvc is consistent
            ctxt = obj.get_context() or ''
            nsvc = f'{obj.get_name()}:{obj.get_stream()}:' \
                   f'{obj.get_version()}:{ctxt}'
            profiles[nsvc] = sorted(obj.get_profiles().keys())
            deplist = []
            for dep in obj.get_dependencies():
                for m, s in dep.peek_requires().items():
                    deplist.append(f'{m}:{",".join(s.get())}'
                                   if len(s.get()) else m)
            deps[nsvc] = sorted(deplist)
        elif isinstance(obj, Modulemd.Defaults):
            module_name = obj.peek_module_name()
            dstreams[module_name] = obj.peek_default_stream()
            for s, pset in obj.peek_profile_defaults().items():
                dprofiles[module_name][s] = pset.get()


def summarize_modules(mfilter=None, yamls=None, urls=None, as_tree=False):
    """
    Load Modulemd objects from each yaml file in the `yamls` list or each url
    in `urls` list and print a summary of the modules found with streams,
    context, version, profiles and dependencies information.
    If no files and no ulrs are passed, local cached metadata is used instead.

    *mfilter*: if present, restricts output to the nsvc supplied. Glob pattern
               can be used, e.g for filtering by stream: '*:master:*'

    *yamls*: list of yaml files to parse. If any, locally cached metadata will
             be ignored

    *urls*: list of repositories to read from. If any, locally cached metadata
            will be ignored.

    *as_tree*: print the summary in tree format, grouping information.
    """

    profiles, dstreams, dprofiles, deps = {}, {}, defaultdict(dict), {}
    if yamls:
        for yaml in yamls:
            try:
                objects = Modulemd.objects_from_file(yaml)
            except gi.repository.GLib.GError as gerror:
                raise ClickException(f"Could not read {yaml}: {gerror}") from gerror
            _parse_objs(objects, profiles, dstreams, dprofiles, deps)
    elif urls:
        for url in urls:
            with tempfile.TemporaryDirectory() as local_path:
                rp = _fetchrepodata.RepoPaths(url, local_path)
                _fetchrepodata._download_metadata_files(rp)
                indexes = _fetchrepodata._read_modules(rp)
                _parse_mmd(indexes, profiles, dstreams, dprofiles, deps)
    else:
        profiles = _repodata.get_modules_profiles_lookup()
        dstreams = _repodata.get_modules_default_streams_lookup()
        dprofiles = _repodata.get_modules_default_profiles_lookup()
        deps = _repodata.get_modules_dependencies_lookup()

    mfilter = mfilter or []
    _print_summary(profiles, dstreams, dprofiles, deps, mfilter, as_tree)
